import logging
from typing import Optional, Union, List

import numpy as np
from scipy.sparse import issparse

from deeptime.markov.tools import estimation as msmest
from ._markov_state_model import MarkovStateModelCollection, MarkovStateModel
from .._base import _MSMBaseEstimator, MembershipsChapmanKolmogorovValidator
from .._transition_counting import TransitionCountModel, TransitionCountEstimator
from ...numeric import is_square_matrix

__all__ = ['MaximumLikelihoodMSM']

log = logging.getLogger(__file__)


class MaximumLikelihoodMSM(_MSMBaseEstimator):
    r"""Maximum likelihood estimator for MSMs (:class:`MarkovStateModel <deeptime.markov.msm.MarkovStateModel>`)
    given discrete trajectories or statistics thereof. This estimator produces instances of MSMs in form of
    MSM collections (:class:`MarkovStateModelCollection`) which contain as many MSMs as there are connected
    sets in the counting. A collection of MSMs per default behaves exactly like an ordinary MSM model on the largest
    connected set. The connected set can be switched, changing the state of the collection to be have like an MSM on
    the selected state subset.

    Implementation according to :footcite:`wu2020variational`.

    Parameters
    ----------
    reversible : bool, optional, default=True
        If true compute reversible MarkovStateModel, else non-reversible MarkovStateModel
    stationary_distribution_constraint : (N,) ndarray, optional, default=None
        Stationary vector on the full set of states. Estimation will be made such the the resulting transition
        matrix has this distribution as an equilibrium distribution. Set probabilities to zero if the states which
        should be excluded from the analysis.
    sparse : bool, optional, default=False
        If true compute count matrix, transition matrix and all derived quantities using sparse matrix algebra.
        In this case python sparse matrices will be returned by the corresponding functions instead of numpy arrays.
        This behavior is suggested for very large numbers of states (e.g. > 4000) because it is likely to be much
        more efficient.
    allow_disconnected : bool, optional, default=False
        If set to true, the resulting transition matrix may have disconnected and transient states, and the
        estimated stationary distribution is only meaningful on the respective connected sets.
    maxiter : int, optional, default=1000000
        Optional parameter with reversible = True, sets the maximum number of iterations before the transition
        matrix estimation method exits.
    maxerr : float, optional, default = 1e-8
        Optional parameter with reversible = True. Convergence tolerance for transition matrix estimation. This
        specifies the maximum change of the Euclidean norm of relative stationary probabilities
        (:math:`x_i = \sum_k x_{ik}`). The relative stationary probability changes
        :math:`e_i = (x_i^{(1)} - x_i^{(2)})/(x_i^{(1)} + x_i^{(2)})` are used in order to track changes in small
        probabilities. The Euclidean norm of the change vector, :math:`|e_i|_2`, is compared to maxerr.
    transition_matrix_tolerance : float, default=1e-8
        The tolerance under which a matrix is still considered a transition matrix (only non-negative elements and
        row sums of 1).
    connectivity_threshold : float, optional, default=0.
        Number of counts required to consider two states connected.
    lagtime : int, optional, default=None
        Optional lagtime that can be provided at estimator level if fitting from timeseries directly.

    References
    ----------
    .. footbibliography::
    """

    def __init__(self, reversible: bool = True, stationary_distribution_constraint: Optional[np.ndarray] = None,
                 sparse: bool = False, allow_disconnected: bool = False, maxiter: int = int(1e6), maxerr: float = 1e-8,
                 connectivity_threshold: float = 0, transition_matrix_tolerance: float = 1e-6, lagtime=None):
        super(MaximumLikelihoodMSM, self).__init__(reversible=reversible, sparse=sparse)

        self.stationary_distribution_constraint = stationary_distribution_constraint
        self.allow_disconnected = allow_disconnected
        self.maxiter = maxiter
        self.maxerr = maxerr
        self.connectivity_threshold = connectivity_threshold
        self.transition_matrix_tolerance = transition_matrix_tolerance
        self.lagtime = lagtime

    @property
    def allow_disconnected(self) -> bool:
        r""" If set to true, the resulting transition matrix may have disconnected and transient states. """
        return self._allow_disconnected

    @allow_disconnected.setter
    def allow_disconnected(self, value: bool):
        self._allow_disconnected = bool(value)

    @property
    def stationary_distribution_constraint(self) -> Optional[np.ndarray]:
        r"""
        The stationary distribution constraint that can either be None (no constraint) or constrains the
        count and transition matrices to states with positive stationary vector entries.

        :getter: Yields the currently configured constraint vector, can be None.
        :setter: Sets a stationary distribution constraint by giving a stationary vector as value. The estimated count-
                 and transition-matrices are restricted to states that have positive entries. In case the vector is not
                 normalized, setting it here implicitly copies and normalizes it.
        :type: ndarray or None
        """
        return self._stationary_distribution_constraint

    @stationary_distribution_constraint.setter
    def stationary_distribution_constraint(self, value: Optional[np.ndarray]):
        if value is not None and (np.any(value < 0) or np.any(value > 1)):
            raise ValueError("not a distribution, contained negative entries and/or entries > 1.")
        if value is not None and np.sum(value) != 1.0:
            # re-normalize if not already normalized
            value = np.copy(value) / np.sum(value)
        self._stationary_distribution_constraint = value

    def fetch_model(self) -> Optional[MarkovStateModelCollection]:
        r"""Yields the most recent :class:`MarkovStateModelCollection` that was estimated.
        Can be None if fit was not called.

        Returns
        -------
        model : MarkovStateModelCollection or None
            The most recent markov state model or None.
        """
        return self._model

    def _fit_connected(self, counts):
        from .. import _transition_matrix as tmat

        if isinstance(counts, np.ndarray):
            if not is_square_matrix(counts) or np.any(counts < 0.):
                raise ValueError("If fitting a count matrix directly, only non-negative square matrices can be used.")
            count_model = TransitionCountModel(counts)
        elif isinstance(counts, TransitionCountModel):
            count_model = counts
        else:
            raise ValueError(f"Unknown type of counts {counts}, only n x n ndarray, TransitionCountModel,"
                             f" or TransitionCountEstimators with a count model are supported.")

        if self.stationary_distribution_constraint is not None:
            if len(self.stationary_distribution_constraint) != count_model.n_states_full:
                raise ValueError(f"Stationary distribution constraint must be defined over full "
                                 f"set of states ({count_model.n_states_full}), but contained "
                                 f"{len(self.stationary_distribution_constraint)} elements.")
            if np.any(self.stationary_distribution_constraint[count_model.state_symbols]) == 0.:
                raise ValueError("The count matrix contains symbols that have no probability in the stationary "
                                 "distribution constraint.")
            if count_model.count_matrix.sum() == 0.0:
                raise ValueError("The set of states with positive stationary probabilities is not visited by the "
                                 "trajectories. A MarkovStateModel reversible with respect to the given stationary"
                                 " vector can not be estimated")

        count_matrix = count_model.count_matrix

        # continue sparse or dense?
        if not self.sparse and issparse(count_matrix):
            # converting count matrices to arrays. As a result the
            # transition matrix and all subsequent properties will be
            # computed using dense arrays and dense matrix algebra.
            count_matrix = count_matrix.toarray()

        # restrict stationary distribution to active set
        if self.stationary_distribution_constraint is None:
            statdist = None
        else:
            statdist = self.stationary_distribution_constraint[count_model.state_symbols]
            statdist /= statdist.sum()  # renormalize

        # Estimate transition matrix
        if self.allow_disconnected:
            P = tmat.estimate_P(count_matrix, reversible=self.reversible, fixed_statdist=statdist,
                                maxiter=self.maxiter, maxerr=self.maxerr)
        else:
            opt_args = {}
            # TODO: non-rev estimate of msmtools does not comply with its own api...
            if statdist is None and self.reversible:
                opt_args['return_statdist'] = True
            P = msmest.transition_matrix(count_matrix, reversible=self.reversible,
                                         mu=statdist, maxiter=self.maxiter,
                                         maxerr=self.maxerr, **opt_args)
        # msmtools returns a tuple for statdist_active=None.
        if isinstance(P, tuple):
            P, statdist = P

        if statdist is None and self.allow_disconnected:
            statdist = tmat.stationary_distribution(P, C=count_matrix)
        return (
            P, statdist, counts
        )

    def _needs_strongly_connected_sets(self):
        return self.reversible and self.stationary_distribution_constraint is None

    def fit_from_counts(self, counts: Union[np.ndarray, TransitionCountEstimator, TransitionCountModel]):
        r""" Fits a model from counts in form of a (n, n) count matrix, a :class:`TransitionCountModel` or an instance
        of `TransitionCountEstimator`, which has been fit on data previously.

        Parameters
        ----------
        counts : (n, n) ndarray or TransitionCountModel or TransitionCountEstimator

        Returns
        -------
        self : MaximumLikelihoodMSM
            Reference to self.
        """
        if isinstance(counts, TransitionCountEstimator):
            if counts.has_model:
                counts = counts.fetch_model()
            else:
                raise ValueError("Can only fit on transition count estimator if the estimator "
                                 "has been fit to data previously.")
        elif isinstance(counts, np.ndarray):
            counts = TransitionCountModel(counts)
        elif isinstance(counts, TransitionCountModel):
            counts = counts
        else:
            raise ValueError("Unknown type of counts argument, can only be one of TransitionCountModel, "
                             "TransitionCountEstimator, (N, N) ndarray. But was: {}".format(type(counts)))
        needs_strong_connectivity = self._needs_strongly_connected_sets()
        if not self.allow_disconnected:
            sets = counts.connected_sets(connectivity_threshold=self.connectivity_threshold,
                                         directed=needs_strong_connectivity)
        else:
            sets = [counts.states]
        transition_matrices = []
        statdists = []
        count_models = []
        for subset in sets:
            try:
                sub_counts = counts.submodel(subset)
                fit_result = self._fit_connected(sub_counts)
                transition_matrices.append(fit_result[0])
                statdists.append(fit_result[1])
                count_models.append(fit_result[2])
            except ValueError as e:
                log.warning(f"Skipping state set {subset} due to error in estimation: {str(e)}.")
        if len(transition_matrices) == 0:
            raise ValueError(f"None of the {'strongly' if needs_strong_connectivity else 'weakly'} "
                             f"connected subsets could be fit to data or the state space decayed into "
                             f"individual states only!")

        self._model = MarkovStateModelCollection(transition_matrices, statdists, reversible=self.reversible,
                                                 count_models=count_models,
                                                 transition_matrix_tolerance=self.transition_matrix_tolerance)
        return self

    def fit_from_discrete_timeseries(self, discrete_timeseries: Union[np.ndarray, List[np.ndarray]],
                                     lagtime: int, count_mode: str = "sliding"):
        r"""Fits a model directly from discrete time series data. This type of data can either be a single
        trajectory in form of a 1d integer numpy array or a list thereof.

        Parameters
        ----------
        discrete_timeseries : ndarray or list of ndarray
            Discrete timeseries data.
        lagtime : int
            The lag time under which to estimate state transitions and ultimately also the transition matrix.
        count_mode : str, default="sliding"
            The count mode to use for estimating transition counts. For maximum-likelihood estimation, the recommended
            choice is "sliding". If the MSM should be used for sampling in a
            :class:`BayesianMSM <deeptime.markov.msm.BayesianMSM>`, the recommended choice is "effective", which yields
            transition counts that are statistically uncorrelated. A description can be found
            in :footcite:`noe2015statistical`.

        Returns
        -------
        self : MaximumLikelihoodMSM
            Reference to self.
        """
        count_model = TransitionCountEstimator(lagtime=lagtime, count_mode=count_mode) \
            .fit(discrete_timeseries).fetch_model()
        return self.fit_from_counts(count_model)

    def fit(self, data, *args, **kw):
        r""" Fits a new markov state model according to data.

        Parameters
        ----------
        data : TransitionCountModel or (n, n) ndarray or discrete timeseries
            Input data, can either be :class:`TransitionCountModel <deeptime.markov.TransitionCountModel>` or
            a 2-dimensional ndarray which is interpreted as count matrix or a discrete timeseries (or a list thereof)
            directly.

            In the case of a timeseries, a lagtime must be provided in the keyword arguments. In this case, also the
            keyword argument "count_mode" can be used, which defaults to "sliding".
            See also :meth:`fit_from_discrete_timeseries`.
        *args
            Dummy parameters for scikit-learn compatibility.
        **kw
            Parameters for scikit-learn compatibility and optionally lagtime if fitting with time series data.

        Returns
        -------
        self : MaximumLikelihoodMSM
            Reference to self.

        See Also
        --------
        TransitionCountModel : Transition count model
        TransitionCountEstimator : Estimating transition count models from data

        Examples
        --------
        This example is demonstrating how to fit a Markov state model collection from data which decomposes into a
        collection of two sets of states with corresponding transition matrices.

        >>> from deeptime.markov.msm import MarkovStateModel  # import MSM
        >>> msm1 = MarkovStateModel([[.7, .3], [.3, .7]])  # create first MSM
        >>> msm2 = MarkovStateModel([[.9, .05, .05], [.3, .6, .1], [.1, .1, .8]])  # create second MSM

        Now, simulate a trajectory where the states of msm2 are shifted by a fixed number `2`, i.e., msm1 describes
        states [0, 1] and msm2 describes states [2, 3, 4] in the generated trajectory.

        >>> traj = np.concatenate([msm1.simulate(1000000), 2 + msm2.simulate(1000000)])  # simulate trajectory

        Given the trajectory, we fit a collection of MSMs:

        >>> model = MaximumLikelihoodMSM(reversible=True).fit(traj, lagtime=1).fetch_model()

        The model behaves like a MSM on the largest connected set, but the behavior can be changed by selecting,
        e.g., the second largest connected set:

        >>> model.state_symbols()
        array([2, 3, 4])
        >>> model.select(1)  # change to second largest connected set
        >>> model.state_symbols()
        array([0, 1])

        And this is all the models contained in the collection:

        >>> model.n_connected_msms
        2

        Alternatively, one can fit with a previously estimated count model (that can be restricted to a subset
        of states):

        >>> counts = TransitionCountEstimator(lagtime=1, count_mode="sliding").fit(traj).fetch_model()
        >>> counts = counts.submodel([0, 1])  # select submodel with state symbols [0, 1]
        >>> msm = MaximumLikelihoodMSM(reversible=True).fit(counts).fetch_model()
        >>> msm.state_symbols()
        array([0, 1])

        And this is the only model in the collection:

        >>> msm.n_connected_msms
        1
        """
        if isinstance(data, (TransitionCountModel, TransitionCountEstimator)):
            return self.fit_from_counts(data)
        elif isinstance(data, np.ndarray) and data.ndim == 2 and data.shape[0] == data.shape[1]:
            return self.fit_from_counts(data)
        else:
            if 'lagtime' not in kw.keys() and self.lagtime is None:
                raise ValueError("To fit directly from a discrete timeseries, a lagtime must be provided!")
            return self.fit_from_discrete_timeseries(data, kw.pop('lagtime', self.lagtime),
                                                     kw.pop("count_mode", "sliding"))

    def chapman_kolmogorov_validator(self, n_metastable_sets: int, mlags,
                                     test_model: Optional[MarkovStateModel] = None):
        r"""Returns a Chapman-Kolmogorov validator based on this estimator and a test model.

        Parameters
        ----------
        n_metastable_sets : int
            Number of metastable sets to project the state space down to.
        mlags : int or range or list
            Multiple of lagtimes of the test_model to test against.
        test_model : MarkovStateModel, optional, default=None
            The model that is tested. If not provided, uses this estimator's encapsulated model.

        Returns
        -------
        validator : markov.MembershipsChapmanKolmogorovValidator
            The validator that can be fit on data.

        Raises
        ------
        AssertionError
            If test_model is None and this estimator has not been :meth:`fit` on data yet or the output model
            was not a discrete output model.
        """
        test_model = self.fetch_model() if test_model is None else test_model
        assert test_model is not None, "We need a test model via argument or an estimator which was already" \
                                       "fit to data."
        assert test_model.has_count_model, "The test model needs to have a count model, i.e., be estimated from data."
        pcca = test_model.pcca(n_metastable_sets)
        reference_lagtime = test_model.count_model.lagtime
        return MLMSMChapmanKolmogorovValidator(test_model, self, pcca.memberships, reference_lagtime, mlags)


def _ck_estimate_model_for_lag(estimator, model, data, lag):
    counting_mode = model.count_model.counting_mode
    counts = TransitionCountEstimator(lag, counting_mode).fit(data).fetch_model().submodel_largest()
    return estimator.fit(counts).fetch_model()


class MLMSMChapmanKolmogorovValidator(MembershipsChapmanKolmogorovValidator):

    def fit(self, data, n_jobs=None, progress=None, **kw):
        return super().fit(data, n_jobs, progress, estimate_model_for_lag=_ck_estimate_model_for_lag, **kw)



r"""
.. currentmodule: deeptime.markov.msm

Maximum-likelihood MSMs (ML-MSM) and Bayesian sampling
------------------------------------------------------
.. autosummary::
    :toctree: generated/
    :template: class_nomodule.rst

    MaximumLikelihoodMSM
    MarkovStateModel
    MarkovStateModelCollection

    BayesianMSM
    BayesianPosterior

Observable operator model MSMs (OOMs)
-------------------------------------
.. autosummary::
    :toctree: generated/
    :template: class_nomodule.rst

    OOMReweightedMSM
    KoopmanReweightedMSM

Augmented markov models (AMMs)
------------------------------
.. autosummary::
    :toctree: generated/
    :template: class_nomodule.rst

    AugmentedMSMEstimator
    AugmentedMSM
    AMMOptimizerState
"""
import logging

from ._markov_state_model import MarkovStateModel, MarkovStateModelCollection
from ._maximum_likelihood_msm import MaximumLikelihoodMSM
from ._bayesian_msm import BayesianMSM, BayesianPosterior
from ._koopman_reweighted_msm import KoopmanReweightedMSM, OOMReweightedMSM
from ._augmented_msm import AugmentedMSMEstimator, AugmentedMSM, AMMOptimizerState

# set up null handler
logging.getLogger(__name__).addHandler(logging.NullHandler())



import random
import warnings
from typing import Optional

import numpy as np

from ..base import EstimatorTransformer
from ._cluster_model import ClusterModel
from . import _clustering_bindings as _bd, metrics

from ..util.parallel import handle_n_jobs


def kmeans_plusplus(data, n_clusters: int, metric: str = 'euclidean', callback=None, seed: int = -1,
                    n_jobs: Optional[int] = None):
    r""" Performs kmeans++ initialization. :footcite:`arthur2006k`

    Parameters
    ----------
    data : np.ndarray
        Input data of shape (T, n_dim).
    n_clusters : int
        The number of cluster centers.
    metric : str, default='euclidean'
        Metric to use during clustering, default evaluates to euclidean metric. For a list of available metrics,
        see the :data:`metric registry <deeptime.clustering.metrics>`.
    callback: callable or None
        used for kmeans++ initialization to indicate progress, called once per assigned center.
    seed : int, optional, default=-1
        The random seed. If non-negative, this fixes the random generator's seed and makes results reproducible.
    n_jobs : int, optional, default=None
        Number of jobs.

    Returns
    -------
    centers : np.ndarray
        An (n_centers, dim)-shaped array with a kmeans++ cluster center initial guess.

    References
    ----------
    .. footbibliography::
    """
    n_jobs = handle_n_jobs(n_jobs)
    metric = metrics[metric]()
    return _bd.kmeans.init_centers_kmpp(data, k=n_clusters, random_seed=seed, n_threads=n_jobs,
                                        callback=callback, metric=metric)


class KMeansModel(ClusterModel):
    r"""The K-means clustering model. Stores all important information which are result of the estimation procedure.
    It can also be used to transform data by assigning each frame to its closest cluster center. For an example
    please see the documentation of the superclass :class:`ClusterModel`.

    Parameters
    ----------
    cluster_centers : (k, d) ndarray
        The d-dimensional cluster centers, length of the array should coincide with :attr:`n_clusters`.
    metric : str
        The metric that was used
    tolerance : float, optional, default=None
        Tolerance which was used as convergence criterium. Defaults to `None` so that clustering models
        can be constructed purely from cluster centers and metric.
    inertias : (t,) ndarray or None, optional, default=None
        Value of the cost function over :code:`t` iterations. Defaults to `None` so that clustering models
        can be constructed purely from cluster centers and metric.
    converged : bool, optional, default=False
        Whether the convergence criterium was met.

    See Also
    --------
    ClusterModel
    KMeans
    MiniBatchKMeans
    """

    def __init__(self, cluster_centers, metric: str, tolerance: Optional[float] = None,
                 inertias: Optional[np.ndarray] = None, converged: bool = False):
        super().__init__(cluster_centers, metric, converged=converged)
        self._inertias = inertias
        self._tolerance = tolerance

    @property
    def tolerance(self):
        """
        The tolerance used as stopping criterion in the kmeans clustering loop. In particular, when
        the relative change in the inertia is smaller than the given tolerance value.

        Returns
        -------
        tolerance : float
            the tolerance
        """
        return self._tolerance

    @property
    def inertia(self) -> Optional[int]:
        r"""Sum of squared distances to assigned centers of training data

        .. math:: \sum_{i=1}^k \sum_{x\in S_i} d(x, \mu_i)^2,

        where :math:`x` are the frames assigned to their respective cluster center :math:`S_i`.

        :type: float or None
        """
        if self._inertias is not None and len(self._inertias) > 0:
            return self._inertias[-1]
        else:
            return None

    @property
    def inertias(self) -> Optional[np.ndarray]:
        r""" Series of inertias over the the iterations of k-means.

        :type: (t, dtype=float) ndarray or None
        """
        return self._inertias

    def score(self, data: np.ndarray, n_jobs: Optional[int] = None) -> float:
        r""" Computes how well the model fits to given data by computing the
        :meth:`inertia <deeptime.clustering.kmeans.KMeansModel.inertia>`.

        Parameters
        ----------
        data : (T, d) ndarray, dtype=float or double
            dataset with T entries and d dimensions
        n_jobs : int, optional, default=None
            number of jobs to use

        Returns
        -------
        score : float
            the inertia
        """
        n_jobs = handle_n_jobs(n_jobs)
        return _bd.kmeans.cost_function(data, self.cluster_centers, n_jobs, metrics[self.metric]())


class KMeans(EstimatorTransformer):
    r"""Clusters the data in a way that minimizes the cost function

    .. math:: C(S) = \sum_{i=1}^{k} \sum_{\mathbf{x}_j \in S_i} \left\| \mathbf{x}_j - \boldsymbol\mu_i \right\|^2

    where :math:`S_i` are clusters with centers of mass :math:`\mu_i` and :math:`\mathbf{x}_j` data points
    associated to their clusters.

    The outcome is very dependent on the initialization, in particular we offer "kmeans++" and "uniform". The latter
    picks initial centers random-uniformly over the provided data set. The former tries to find an initialization
    which is covering the spatial configuration of the dataset more or less uniformly. For details
    see :footcite:`arthur2006k`.

    Parameters
    ----------
    n_clusters : int
        amount of cluster centers.
    max_iter : int, default=500
        maximum number of iterations before stopping.
    metric : str, default='euclidean'
        Metric to use during clustering, default evaluates to euclidean metric. For a list of available metrics,
        see the :data:`metric registry <deeptime.clustering.metrics>`.
    tolerance : float, default=1e-5
        Stop iteration when the relative change in the cost function (inertia)

        .. math:: C(S) = \sum_{i=1}^{k} \sum_{\mathbf x \in S_i} \left\| \mathbf x - \boldsymbol\mu_i \right\|^2

        is smaller than tolerance.
    init_strategy : str, default='kmeans++'
        one of 'kmeans++', 'uniform'; determining how the initial cluster centers are being chosen
    fixed_seed : bool or int, default=False
        if True, the seed gets set to 42. Use time based seeding otherwise. If an integer is given, use this to
        initialize the random generator.
    n_jobs : int or None, default=None
        Number of threads to use during clustering and assignment of data. If None, one core will be used.
    initial_centers: None or np.ndarray[k, dim], default=None
        This is used to resume the kmeans iteration. Note, that if this is set, the init_strategy is ignored and
        the centers are directly passed to the kmeans iteration algorithm.

    References
    ----------
    .. footbibliography::

    See Also
    --------
    KMeansModel
    MiniBatchKMeans
    """

    def __init__(self, n_clusters: int, max_iter: int = 500, metric='euclidean',
                 tolerance=1e-5, init_strategy: str = 'kmeans++', fixed_seed=False,
                 n_jobs=None, initial_centers=None):
        super(KMeans, self).__init__()

        self.n_clusters = n_clusters
        self.max_iter = max_iter
        self.metric = metric
        self.tolerance = tolerance
        self.init_strategy = init_strategy
        self.fixed_seed = fixed_seed
        self.random_state = np.random.RandomState(self.fixed_seed)
        self.n_jobs = handle_n_jobs(n_jobs)
        self.initial_centers = initial_centers

    @property
    def initial_centers(self) -> Optional[np.ndarray]:
        r"""
        Yields initial centers which override the :meth:`init_strategy`. Can be used to resume k-means iterations.

        :getter: The initial centers or None.
        :setter: Sets the initial centers. If not None, the array is expected to have length :attr:`n_clusters`.
        :type: (k, n) ndarray or None
        """
        return self._initial_centers

    @initial_centers.setter
    def initial_centers(self, value: Optional[np.ndarray]):
        if value is not None and value.shape[0] != self.n_clusters:
            raise ValueError("initial centers must be None or of shape (k, d) where k is the number of cluster centers."
                             " Expected k={}, got {}.".format(self.n_clusters, value.shape[0]))
        self._initial_centers = value

    @property
    def n_jobs(self) -> int:
        r"""
        Number of threads to use during clustering and assignment of data.

        :getter: Yields the number of threads. If -1, all available threads are used.
        :setter: Sets the number of threads to use. If -1, use all, if None, use 1.
        :type: int
        """
        return self._n_jobs

    @n_jobs.setter
    def n_jobs(self, value: Optional[int]):
        self._n_jobs = handle_n_jobs(value)

    @property
    def n_clusters(self) -> int:
        r"""
        The number of cluster centers to use.

        :getter: Yields the number of cluster centers.
        :setter: Sets the number of cluster centers.
        :type: int
        """
        return self._n_clusters

    @n_clusters.setter
    def n_clusters(self, value: int):
        self._n_clusters = value

    @property
    def max_iter(self) -> int:
        r"""
        Maximum number of clustering iterations before stop.

        :getter: Yields the maximum number of clustering iterations
        :setter: Sets the max. number of clustering iterations
        :type: int
        """
        return self._max_iter

    @max_iter.setter
    def max_iter(self, value: int):
        self._max_iter = value

    @property
    def tolerance(self) -> float:
        r"""
        Stopping criterion for the k-means iteration. When the relative change of the cost function between two
        iterations is less than the tolerance, the algorithm is considered to be converged.

        :getter: Yields the currently set tolerance.
        :setter: Sets a new tolerance.
        :type: float
        """
        return self._tolerance

    @tolerance.setter
    def tolerance(self, value: float):
        self._tolerance = value

    @property
    def metric(self) -> str:
        r"""
        The metric that is used for clustering.

        See Also
        --------
        _clustering_bindings.Metric : The metric class, can be subclassed
        metrics : Metrics registry which maps from metric label to actual implementation
        """
        return self._metric

    @metric.setter
    def metric(self, value: str):
        if value not in metrics.available:
            raise ValueError(f"Unknown metric {value}, available metrics: {metrics.available}")
        self._metric = value

    def fetch_model(self) -> Optional[KMeansModel]:
        """
        Fetches the current model. Can be `None` in case :meth:`fit` was not called yet.

        Returns
        -------
        model : KMeansModel or None
            the latest estimated model
        """
        return self._model

    def transform(self, data, **kw) -> np.ndarray:
        """
        Transforms a trajectory to a discrete trajectory by assigning each frame to its respective cluster center.

        Parameters
        ----------
        data : (T, n) ndarray
            trajectory with `T` frames and data points in `n` dimensions.
        **kw
            ignored kwargs for scikit-learn compatibility

        Returns
        -------
        discrete_trajectory : (T, 1) ndarray
            discrete trajectory

        See Also
        --------
        ClusterModel.transform : transform method of cluster model, implicitly called.
        """
        return super().transform(data)

    @property
    def init_strategy(self):
        r"""Strategy to get an initial guess for the centers.

        :getter: Yields the strategy, can be one of "kmeans++" or "uniform".
        :setter: Setter for the initialization strategy that is used when no initial centers are provided.
        :type: string
        """
        return self._init_strategy

    @init_strategy.setter
    def init_strategy(self, value: str):
        valid = ('kmeans++', 'uniform')
        if value not in valid:
            raise ValueError('invalid parameter "{}" for init_strategy. Should be one of {}'.format(value, valid))
        self._init_strategy = value

    @property
    def fixed_seed(self):
        """ seed for random choice of initial cluster centers.

        Fix this to get reproducible results in conjunction with n_jobs=0. The latter is needed, because parallel
        execution causes non-deterministic behaviour again.
        """
        return self._fixed_seed

    @fixed_seed.setter
    def fixed_seed(self, value: [bool, int, None]):
        """
        Sets a fixed seed for cluster estimation to get reproducible results. This only works in the case of n_jobs=0
        or n_jobs=1, parallel execution reintroduces non-deterministic behavior.

        Parameters
        ----------
        value : bool or int or None
            If the value is `True`, the seed will be fixed on `42`, if it is `False` or `None`, the seed gets drawn
            randomly. In case an `int` value is provided, that will be used as fixed seed.

        """
        if isinstance(value, bool) or value is None:
            if value:
                self._fixed_seed = 42
            else:
                self._fixed_seed = random.randint(0, 2 ** 32 - 1)
        elif isinstance(value, int):
            if value < 0 or value > 2 ** 32 - 1:
                warnings.warn("seed has to be non-negative (or smaller than 2**32)."
                              " Seed will be chosen randomly.")
                self.fixed_seed = False
            else:
                self._fixed_seed = value
        else:
            raise ValueError("fixed seed has to be None, bool or integer")

    def _pick_initial_centers(self, data, strategy, n_jobs, callback=None):
        if self.n_clusters > len(data):
            raise ValueError('Not enough data points for desired amount of clusters.')

        if strategy == 'uniform':
            return data[self.random_state.randint(0, len(data), size=self.n_clusters)]
        elif self.init_strategy == 'kmeans++':
            return kmeans_plusplus(data, self.n_clusters, self.metric,
                                   callback=callback, seed=self.fixed_seed, n_jobs=n_jobs)

    def fit(self, data, initial_centers=None, callback_init_centers=None, callback_loop=None, n_jobs=None):
        """ Perform the clustering.

        Parameters
        ----------
        data: np.ndarray
            data to be clustered, shape should be (N, D), where N is the number of data points, D the dimension.
            In case of one-dimensional data, a shape of (N,) also works.
        initial_centers: np.ndarray or None
            Optional cluster center initialization that supersedes the estimator's `initial_centers` attribute
        callback_init_centers: function or None
            used for kmeans++ initialization to indicate progress, called once per assigned center.
        callback_loop: function or None
            used to indicate progress on kmeans iterations, called once per iteration.
        n_jobs: None or int
            if not None, supersedes the n_jobs attribute of the estimator instance; must be non-negative

        Returns
        -------
        self : KMeans
            reference to self
        """
        if data.ndim == 1:
            data = data[:, np.newaxis]
        n_jobs = handle_n_jobs(self.n_jobs if n_jobs is None else n_jobs)
        if initial_centers is not None:
            self.initial_centers = initial_centers
        if self.initial_centers is None:
            self.initial_centers = self._pick_initial_centers(data, self.init_strategy, n_jobs, callback_init_centers)

        # run k-means with all the data
        converged = False
        cluster_centers, code, iterations, cost = _bd.kmeans.cluster_loop(
            data, self.initial_centers.copy(), n_jobs, self.max_iter,
            self.tolerance, callback_loop, metrics[self.metric]())
        if code == 0:
            converged = True
        else:
            warnings.warn(f"Algorithm did not reach convergence criterion"
                          f" of {self.tolerance} in {self.max_iter} iterations. Consider increasing max_iter.")
        self._model = KMeansModel(metric=self.metric, tolerance=self.tolerance,
                                  cluster_centers=cluster_centers, inertias=cost, converged=converged)

        return self


class MiniBatchKMeans(KMeans):
    r""" K-means clustering in a mini-batched fashion.

    Parameters
    ----------
    batch_size : int, optional, default=100
        The maximum sample size if calling :meth:`fit()`.

    See Also
    --------
    KMeans : Superclass, see for description of remaining parameters.
    KMeansModel
    """

    def __init__(self, n_clusters, batch_size=100, max_iter=5, metric='euclidean', tolerance=1e-5,
                 init_strategy='kmeans++', n_jobs=None, initial_centers=None):
        super(MiniBatchKMeans, self).__init__(n_clusters, max_iter, metric,
                                              tolerance, init_strategy, False,
                                              n_jobs=n_jobs,
                                              initial_centers=initial_centers)
        self.batch_size = batch_size

    def fit(self, data, initial_centers=None, callback_init_centers=None, callback_loop=None, n_jobs=None):
        r""" Perform clustering on whole data. """
        n_jobs = handle_n_jobs(self.n_jobs if n_jobs is None else n_jobs)
        if data.ndim == 1:
            data = data[:, np.newaxis]
        if initial_centers is not None:
            self.initial_centers = initial_centers
        if self.initial_centers is None:
            self.initial_centers = self._pick_initial_centers(data, self.init_strategy, n_jobs, callback_init_centers)
        indices = np.arange(len(data))
        self._model = KMeansModel(cluster_centers=self.initial_centers, metric=self.metric,
                                  tolerance=self.tolerance, inertias=np.array([float('inf')]))
        for epoch in range(self.max_iter):
            np.random.shuffle(indices)
            if len(data) > self.batch_size:
                split = np.array_split(indices, len(data) // self.batch_size)
            else:
                split = [indices]
            for chunk_indices in split:
                self.partial_fit(data[chunk_indices], n_jobs=n_jobs)
                if self._model.converged:
                    break
        return self

    def partial_fit(self, data, n_jobs=None):
        r"""
        Updates the current model (or creates a new one) with data. This method can be called repeatedly and thus
        be used to train a model in an on-line fashion. Note that usually multiple passes over the data is used.
        Also this method should not be mixed with calls to :meth:`fit`, as then the model is overwritten with a new
        instance based on the data passed to :meth:`fit`.

        Parameters
        ----------
        data : (T, n) ndarray
            Data with which the model is updated and/or initialized.
        n_jobs : int, optional, default=None
            number of jobs to use when updating the model, supersedes the n_jobs attribute of the estimator.

        Returns
        -------
        self : MiniBatchKMeans
            reference to self
        """
        n_jobs = handle_n_jobs(self.n_jobs if n_jobs is None else n_jobs)
        if self.initial_centers is None:
            # we have no initial centers set, pick some based on the first partial fit
            self.initial_centers = self._pick_initial_centers(data, self.init_strategy, n_jobs)

        if self._model is None:
            self._model = KMeansModel(cluster_centers=np.copy(self.initial_centers), metric=self.metric,
                                      tolerance=self.tolerance, inertias=np.array([float('inf')]))
        if data.ndim == 1:
            data = data[:, np.newaxis]
        metric_instance = metrics[self.metric]()
        self._model._cluster_centers = _bd.kmeans.cluster(data, self._model.cluster_centers, n_jobs, metric_instance)[0]
        cost = _bd.kmeans.cost_function(data, self._model.cluster_centers, n_jobs, metric_instance)

        rel_change = np.abs(cost - self._model.inertia) / cost if cost != 0.0 else 0.0
        self._model._inertias = np.append(self._model._inertias, cost)

        if rel_change <= self.tolerance:
            self._model._converged = True

        return self



r"""
.. currentmodule: deeptime.data

===============================================================================
API
===============================================================================

.. autosummary::
    :toctree: generated/
    :template: class_nomodule.rst

    double_well_discrete
    quadruple_well
    quadruple_well_asymmetric
    triple_well_2d
    triple_well_1d
    double_well_2d
    time_dependent_quintuple_well
    abc_flow
    ornstein_uhlenbeck
    prinz_potential
    ellipsoids
    sqrt_model
    swissroll_model
    position_based_fluids
    drunkards_walk
    bickley_jet
    birth_death_chain
    tmatrix_metropolis1d

===============================================================================
Custom systems
===============================================================================

.. autosummary::
    :toctree: generated/
    :template: class_nomodule.rst

    custom_sde
    custom_ode

===============================================================================
Implementations
===============================================================================

.. autosummary::
    :toctree: generated/impl/
    :template: class_nomodule.rst

    DoubleWellDiscrete
    Ellipsoids
    PBFSimulator
    DrunkardsWalk
    BickleyJet
    BirthDeathChain

    TimeDependentSystem
    TimeIndependentSystem
    CustomSystem
"""

from ._systems import CustomSystem, TimeIndependentSystem, TimeDependentSystem
from ._datasets import double_well_discrete, ellipsoids, position_based_fluids, drunkards_walk, bickley_jet, \
    birth_death_chain, tmatrix_metropolis1d, sqrt_model, quadruple_well, triple_well_2d, abc_flow, ornstein_uhlenbeck, \
    triple_well_1d, quadruple_well_asymmetric, double_well_2d, swissroll_model, prinz_potential, \
    time_dependent_quintuple_well
from ._datasets import custom_sde, custom_ode
from ._double_well import DoubleWellDiscrete
from ._ellipsoids import Ellipsoids
from ._pbf_simulator import PBFSimulator
from ._drunkards_walk_simulator import DrunkardsWalk
from ._bickley_simulator import BickleyJet
from ._birth_death_chain import BirthDeathChain



from typing import Optional, Callable

import numpy as np

from deeptime.basis import Identity
from . import TransferOperatorModel
from ..base import EstimatorTransformer
from ..covariance import Covariance
from ..kernels import Kernel
from ..numeric import spd_truncated_svd
from ..util.types import to_dataset


class _KVADTransform:

    def __init__(self, cov, obs_transform, singular_vectors):
        self.cov = cov
        self.obs_transform = obs_transform
        self.singular_vectors = singular_vectors

    def __call__(self, x):
        return self.cov.whiten(self.obs_transform(x)) @ self.singular_vectors


class KVADModel(TransferOperatorModel):
    r"""The model produced by the :class:`KVAD` estimator.

    Parameters
    ----------
    kernel : Kernel
        The kernel that was used for estimation and embedding.
    koopman_matrix : ndarray
        The estimated Koopman matrix.
    observable_transform : callable
        Transformation for data that was used for estimation.
    covariances : deeptime.covariance.CovarianceModel
        Estimated covariances for instantaneous data.
    singular_values : ndarray
        Singular values of truncated SVD.
    singular_vectors : ndarray
        Singular vectors of truncated SVD.
    score : float
        Estimated KVAD score.
    """

    def __init__(self, kernel, koopman_matrix: np.ndarray, observable_transform, covariances,
                 singular_values, singular_vectors, score):
        transf = _KVADTransform(covariances, observable_transform, singular_vectors)
        super(KVADModel, self).__init__(koopman_matrix=koopman_matrix,
                                        instantaneous_obs=transf,
                                        timelagged_obs=transf)
        self.kernel = kernel
        self.observable_transform = observable_transform
        self.covariances = covariances
        self.singular_values = singular_values
        self.singular_vectors = singular_vectors
        self.score = score


class KVAD(EstimatorTransformer):
    r""" An estimator for the "Kernel embedding based variational approach for dynamical systems" (KVAD).

    Theory and introduction into the method can be found in :footcite:`tian2020kernel`.

    Parameters
    ----------
    kernel : Kernel
        The kernel to be used, see :mod:`deeptime.kernels` for a selection of predefined kernels.
    lagtime : int, optional, default=None
        Lagtime if data is not a list of instantaneous and time-lagged data pairs but a trajectory instead.
    dim : int, optional, default=None
        Dimension cutoff parameter.
    epsilon : float, default=1e-6
        Regularization parameter for truncated SVD.
    observable_transform : callable, optional, default=Identity
        A feature transformation on the raw data which is used to estimate the model.

    See Also
    --------
    KVADModel

    References
    ----------
    .. footbibliography::
    """

    def __init__(self, kernel: Kernel, lagtime: Optional[int] = None,
                 dim: Optional[int] = None,
                 epsilon: float = 1e-6,
                 observable_transform: Callable[[np.ndarray], np.ndarray] = Identity()):
        super().__init__()
        self.kernel = kernel
        self.dim = dim
        self.lagtime = lagtime
        self.epsilon = epsilon
        self.observable_transform = observable_transform

    @property
    def observable_transform(self) -> Callable[[np.ndarray], np.ndarray]:
        r""" Transforms observable instantaneous and time-lagged data into feature space.

        :type: Callable[[ndarray], ndarray]
        """
        return self._observable_transform

    @observable_transform.setter
    def observable_transform(self, value: Callable[[np.ndarray], np.ndarray]):
        self._observable_transform = value

    @property
    def epsilon(self):
        r""" Regularization parameter for truncated SVD.

        :type: float
        """
        return self._epsilon

    @epsilon.setter
    def epsilon(self, value):
        self._epsilon = value

    @property
    def dim(self) -> Optional[int]:
        r""" Dimension cutoff for the decomposition.

        :type: int or None
        """
        return self._dim

    @dim.setter
    def dim(self, value: Optional[int]):
        self._dim = value

    def fit(self, data, **kwargs):
        dataset = to_dataset(data, lagtime=self.lagtime)
        x, y = dataset[:]

        chi_x = self.observable_transform(x)
        chi_y = self.observable_transform(y)

        n_data = y.shape[0]
        chi_output_dim = chi_x.shape[1]

        assert chi_x.shape == (n_data, chi_output_dim)
        assert chi_y.shape == (n_data, chi_output_dim)

        g_yy = self.kernel.gram(y)
        assert g_yy.shape == (n_data, n_data)

        cov = Covariance().fit(chi_x).fetch_model()
        chi_x_w = cov.whiten(chi_x, epsilon=self.epsilon)
        chi_y_w = cov.whiten(chi_y, epsilon=self.epsilon)

        x_g_x = np.linalg.multi_dot((chi_x_w.T, g_yy, chi_x_w)) / (n_data * n_data)
        singular_values, singular_vectors = spd_truncated_svd(x_g_x, dim=self.dim, eps=self.epsilon)

        f_x = chi_x_w @ singular_vectors
        f_y = chi_y_w @ singular_vectors

        koopman_matrix = np.zeros((len(singular_values) + 1, len(singular_values) + 1), dtype=y.dtype)
        koopman_matrix[0, 0] = 1
        koopman_matrix[0, 1:] = singular_vectors.T.dot(chi_y_w.mean(axis=0))
        koopman_matrix[1:, 1:] = (1. / n_data) * f_x.T @ f_y

        score = np.sum(singular_values) + np.mean(g_yy)
        self._model = KVADModel(self.kernel, koopman_matrix, self.observable_transform,
                                cov, singular_values, singular_vectors, score)
        return self



import numpy as np


def schatten_norm(mat: np.ndarray, order: float = 1, hermitian: bool = False) -> float:
    r""" Computes the r-Schatten norm

    .. math::
        \| T \|_r = \left( \sum_i \sigma_i^p (T) \right)^{1/r},

    where :math:`\sigma_i` are the singular values corresponding to matrix :math:`T`.

    Parameters
    ----------
    mat : ndarray
        input matrix
    order : float
        Order of the Schatten norm, must be :math:`\geq 1`.
    hermitian : bool, optional, default=False
        If True, the matrix is assumed to be Hermitian, which allows the norm to be computed more efficiently.

    Returns
    -------
    norm : float
        The norm.
    """
    assert order >= 1, 'Order only defined for r >= 1'
    if order == 1:
        return np.linalg.norm(mat, ord='nuc')
    elif order == 2:
        return np.linalg.norm(mat, ord='fro')
    else:
        s = np.linalg.svd(mat, compute_uv=False, hermitian=hermitian)
        return np.power(np.sum(np.power(s, order)), 1. / order)



import abc
from collections import defaultdict
from inspect import signature
from typing import Optional

from sklearn.base import _pprint as pprint_sklearn


class _BaseMethodsMixin(abc.ABC):
    """ Defines common methods used by both Estimator and Model classes. These are mostly static and low-level
    checking of conformity with respect to deeptime conventions.
    """

    def __repr__(self):
        name = '{cls}-{id}:'.format(id=id(self), cls=self.__class__.__name__)
        return '{name}{params}]'.format(
            name=name, params=pprint_sklearn(self.get_params(), offset=len(name), )
        )

    def get_params(self, deep=False):
        r"""Get the parameters.

        Returns
        -------
        params : mapping of string to any
            Parameter names mapped to their values.
        """
        params = dict()

        # introspect the constructor arguments to find the model parameters
        # to represent
        cls = self.__class__
        init = getattr(cls.__init__, 'deprecated_original', cls.__init__)
        init_sign = signature(init)
        args, varargs = [], []
        for parameter in init_sign.parameters.values():
            if (parameter.kind != parameter.VAR_KEYWORD and
                    parameter.name != 'self'):
                args.append(parameter.name)
            if parameter.kind == parameter.VAR_POSITIONAL:
                varargs.append(parameter.name)

        if len(varargs) != 0:
            raise RuntimeError("scikit-learn kernels should always "
                               "specify their parameters in the signature"
                               " of their __init__ (no varargs)."
                               " %s doesn't follow this convention."
                               % (cls,))
        for arg in args:
            params[arg] = getattr(self, arg, None)
        return params

    def set_params(self, **params):
        """
        Set the parameters of this estimator.

        The method works on simple estimators as well as on nested objects
        (such as pipelines). The latter have parameters of the form
        ``<component>__<parameter>`` so that it's possible to update each
        component of a nested object.

        Parameters
        ----------
        **params : dict
            Estimator parameters.

        Returns
        -------
        self : object
            Estimator instance.
        """
        if not params:
            # Simple optimization to gain speed (inspect is slow)
            return self
        valid_params = self.get_params(deep=True)

        nested_params = defaultdict(dict)  # grouped by prefix
        for key, value in params.items():
            key, delim, sub_key = key.partition('__')
            if key not in valid_params:
                raise ValueError('Invalid parameter %s for estimator %s. '
                                 'Check the list of available parameters '
                                 'with `estimator.get_params().keys()`.' %
                                 (key, self))

            if delim:
                nested_params[key][sub_key] = value
            else:
                setattr(self, key, value)
                valid_params[key] = value

        for key, sub_params in nested_params.items():
            valid_params[key].set_params(**sub_params)

        return self

    def __getstate__(self):
        try:
            state = super().__getstate__()
        except AttributeError:
            state = self.__dict__

        if type(self).__module__.startswith('deeptime.'):
            from deeptime import __version__
            return dict(state.items(), _deeptime_version=__version__)
        else:
            return state

    def __setstate__(self, state):
        from deeptime import __version__
        if type(self).__module__.startswith('deeptime.'):
            pickle_version = state.pop("_deeptime_version", None)
            if pickle_version != __version__:
                import warnings
                warnings.warn(
                    "Trying to unpickle estimator {0} from version {1} when "
                    "using version {2}. This might lead to breaking code or "
                    "invalid results. Use at your own risk.".format(
                        self.__class__.__name__, pickle_version, __version__),
                    UserWarning)
        try:
            super().__setstate__(state)
        except AttributeError:
            self.__dict__.update(state)


class Dataset(abc.ABC):
    r""" The Dataset superclass. It is an abstract class requiring implementations of

    * :meth:`__len__` to obtain its length,
    * :meth:`__getitem__` to obtain one or several items,
    * :meth:`setflags` to set its writeable state.

    See Also
    --------
    deeptime.util.data.TimeLaggedDataset
        A dataset implementation for pairs of instantaneous and timelagged data.
    deeptime.util.data.TimeLaggedConcatDataset
        A concatenation of several :class:`TimeLaggedDataset <deeptime.util.data.TimeLaggedDataset>` .
    deeptime.util.data.TrajectoryDataset
        A dataset for one trajectory with a lagtime.
    deeptime.util.data.TrajectoriesDataset
        A dataset for multiple trajectories with a lagtime.
    """

    @abc.abstractmethod
    def setflags(self, write=True):
        r""" Set writeable flags for contained arrays. """

    @abc.abstractmethod
    def __len__(self):
        r""" Length of this dataset. """

    @abc.abstractmethod
    def __getitem__(self, item):
        r""" Retrieves one or multiple items from this dataset. """


class Model(_BaseMethodsMixin):
    r""" The model superclass. """

    def copy(self) -> "Model":
        r""" Makes a deep copy of this model.

        Returns
        -------
        copy
            A new copy of this model.
        """
        import copy
        return copy.deepcopy(self)


class Estimator(_BaseMethodsMixin):
    r""" Base class of all estimators

    Parameters
    ----------
    model : Model, optional, default=None
        A model which can be used for initialization. In case an estimator is capable of online learning, i.e.,
        capable of updating models, this can be used to resume the estimation process.
    """

    """ class wide flag to control whether input of fit or partial_fit should be checked for modifications """
    _MUTABLE_INPUT_DATA = False

    def __init__(self, model=None):
        self._model = model

    @abc.abstractmethod
    def fit(self, data, **kwargs):
        r""" Fits data to the estimator's internal :class:`Model` and overwrites it. This way, every call to
        :meth:`fetch_model` yields an autonomous model instance. Sometimes a :code:`partial_fit` method is available,
        in which case the model can get updated by the estimator.

        Parameters
        ----------
        data : array_like
            Data that is used to fit a model.
        **kwargs
            Additional kwargs.

        Returns
        -------
        self : Estimator
            Reference to self.
        """

    def fetch_model(self) -> Optional[Model]:
        r""" Yields the estimated model. Can be None if :meth:`fit` was not called.

        Returns
        -------
        model : Model or None
            The estimated model or None.
        """
        return self._model

    def fit_fetch(self, data, **kwargs):
        r""" Fits the internal model on data and subsequently fetches it in one call.

        Parameters
        ----------
        data : array_like
            Data that is used to fit the model.
        **kwargs
            Additional arguments to :meth:`fit`.

        Returns
        -------
        model
            The estimated model.
        """
        self.fit(data, **kwargs)
        return self.fetch_model()

    @property
    def model(self):
        """ Shortcut to :meth:`fetch_model`. """
        return self.fetch_model()

    @property
    def has_model(self) -> bool:
        r""" Property reporting whether this estimator contains an estimated model. This assumes that the model
        is initialized with `None` otherwise.

        :type: bool
        """
        return self._model is not None

    def __getattribute__(self, item):
        if (item == 'fit' or item == 'partial_fit') and not self._MUTABLE_INPUT_DATA:
            fit = super(Estimator, self).__getattribute__(item)
            return _ImmutableInputData(fit)

        return super(_BaseMethodsMixin, self).__getattribute__(item)


class _ImmutableInputData:
    """A function decorator for Estimator.fit() to make input data immutable """

    def __init__(self, fit_method):
        self.fit_method = fit_method
        self._data = None

    @property
    def data(self):
        return self._data

    @data.setter
    def data(self, value_):
        import numpy as np
        args, kwargs = value_
        self._data = []

        # first argument is x
        if len(args) == 0:
            if 'data' in kwargs:
                args = [kwargs['data']]
            elif len(kwargs) == 1:
                args = [kwargs[k] for k in kwargs.keys()]
            else:
                raise InputFormatError(f'No input at all for fit(). Input was {args}, kw={kwargs}')
        value = args[0]
        if hasattr(value, 'setflags'):
            self._data.append(value)
        elif isinstance(value, (list, tuple)):
            for i, x in enumerate(value):
                if isinstance(x, np.ndarray):
                    self._data.append(x)
                else:
                    raise InputFormatError(f'Invalid input element in position {i}, only numpy.ndarrays allowed.')
        elif isinstance(value, (Model, Estimator)):
            self._data.append(value)
        else:
            raise InputFormatError(f'Only estimator, model, ndarray or list/tuple of ndarray allowed. '
                                   f'But was of type {type(value)}: {value}.')

    def __enter__(self):
        import numpy as np
        self.old_writable_flags = []
        for d in self.data:
            if isinstance(d, np.ndarray):
                self.old_writable_flags.append(d.flags.writeable)
                # set ndarray writabe flags to false
                try:
                    d.setflags(write=False)
                except:
                    # although this should not raise, occasionally it does raise
                    # due to arrays stemming from torch which then cannot be set immutable
                    ...

    def __exit__(self, exc_type, exc_val, exc_tb):
        # restore ndarray writable flags to old state
        import numpy as np
        for d, writable in zip(self.data, self.old_writable_flags):
            if isinstance(d, np.ndarray):
                try:
                    d.setflags(write=writable)
                except:
                    # although this should not raise, occasionally it does raise
                    # due to arrays stemming from torch which then cannot be set immutable
                    ...

    def __call__(self, *args, **kwargs):
        # extract input data from args, **kwargs (namely x and y)
        self.data = args, kwargs

        # here we invoke the immutable setting context manager.
        with self:
            return self.fit_method(*args, **kwargs)


class Transformer(abc.ABC):
    r""" Base class of all transformers. """

    @abc.abstractmethod
    def transform(self, data, **kwargs):
        r"""Transforms the input data.

        Parameters
        ----------
        data : array_like
            Input data.

        Returns
        -------
        transformed : array_like
            The transformed data
        """

    def __call__(self, *args, **kwargs):
        return self.transform(*args, **kwargs)


class EstimatorTransformer(Estimator, Transformer, abc.ABC):

    def fit_transform(self, data, fit_options=None, transform_options=None):
        r""" Fits a model which simultaneously functions as transformer and subsequently transforms
        the input data. The estimated model can be accessed by calling :meth:`fetch_model`.

        Parameters
        ----------
        data : array_like
            The input data.
        fit_options : dict, optional, default=None
            Optional keyword arguments passed on to the fit method.
        transform_options : dict, optional, default=None
            Optional keyword arguments passed on to the transform method.

        Returns
        -------
        output : array_like
            Transformed data.
        """
        fit_options = {} if fit_options is None else fit_options
        transform_options = {} if transform_options is None else transform_options
        return self.fit(data, **fit_options).transform(data, **transform_options)

    def transform(self, data, **kwargs):
        r""" Transforms data with the encapsulated model.

        Parameters
        ----------
        data : array_like
            Input data
        **kwargs
            Optional arguments.

        Returns
        -------
        output : array_like
            Transformed data.
        """
        model = self.fetch_model()
        if model is None:
            raise ValueError("This estimator contains no model yet, fit should be called first.")
        return model.transform(data, **kwargs)


class InputFormatError(ValueError):
    """Input data for Estimator is not allowed."""



from typing import List

import numpy as np

from ._basis_bindings import evaluate_monomials as _eval, power_matrix as _power_matrix, feature_names as _fn
from ._base import Observable


class Identity(Observable):
    r""" The identity. """

    def _evaluate(self, x):
        return x

    @staticmethod
    def get_feature_names(input_features=None):
        if input_features is None:
            return ['x']
        else:
            assert len(input_features) == 1
            return input_features


class Monomials(Observable):
    r""" Monomial basis observable which transforms a number of d-dimensional
    datapoints :math:`\mathbf{x}\in\mathbb{R}^d` into (unique) monomials of at most degree :math:`p`.

    This means, that

    .. math::

        \mathbf{x} \mapsto \left\{ \prod_{d=1}^n \mathbf{x}_d^{k_d} : \sum k_d \leq p \right\}.

    The set is returned as a numpy ndarray of shape `(n_test_points, n_monomials)`, where `n_monomials` is the
    size of the set.

    Parameters
    ----------
    p : int
        Maximum degree of the monomial basis. Must be positive.
    d : int
        The dimension of the input.

    Examples
    --------
    Given three test points in one dimension

    >>> import numpy as np
    >>> X = np.random.normal(size=(3, 1))

    Evaluating the monomial basis up to degree two yields :math:`x^0, x^1, x^2`, i.e., the expected shape is (3, 3)

    >>> Y = Monomials(p=2, d=1)(X)
    >>> Y.shape
    (3, 3)

    and, e.g., the second monomial of the third test point is the third test point itself:

    >>> np.testing.assert_almost_equal(Y[2, 1], X[2, 0])
    """

    def __init__(self, p: int, d: int):
        from scipy.special import binom

        assert p > 0
        self.p = p
        self.d = d
        self._n_monomials = int(binom(self.p + self.d, self.p))
        self._power_matrix = _power_matrix(self.d, self._n_monomials)

    def _evaluate(self, x: np.ndarray):
        if self.d is not None and x.shape[1] != self.d:
            raise ValueError(f"Input had the wrong dimension {x.shape[1]}, this basis requires {self.d}.")
        return _eval(self.p, x.T, self._power_matrix).T

    def get_feature_names(self, input_features=None) -> List[str]:
        r""" Yields a list of feature names, optionally given input feature names.

        Parameters
        ----------
        input_features : list of str, optional, default=None
            If not `None`, replaces the input feature names.

        Returns
        -------
        feature_names : list of str
            Feature names corresponding to each monomial.
        """
        if input_features is None:
            input_features = [f'x{i}' for i in range(self.d)]
        return _fn(input_features, self._power_matrix)



def configuration(parent_package='', top_path=None):
    from numpy.distutils.misc_util import Configuration

    config = Configuration('deeptime', parent_package, top_path)

    config.add_subpackage('clustering')
    config.add_subpackage('covariance')
    config.add_subpackage('data')
    config.add_subpackage('decomposition')
    config.add_subpackage('markov')
    config.add_subpackage('numeric')
    config.add_subpackage('kernels')
    config.add_subpackage('basis')
    config.add_subpackage('util')
    config.add_subpackage('sindy')

    from Cython.Build import cythonize
    config.ext_modules = cythonize(
        config.ext_modules,
        compiler_directives={'language_level': '3'})

    return config



r"""This module implements effective transition counts

.. moduleauthor:: F. Noe <frank DOT noe AT fu-berlin DOT de>

"""

import numpy as np
import scipy.sparse
from scipy.sparse.csr import csr_matrix

from threadpoolctl import threadpool_limits

from deeptime.util.stats import statistical_inefficiency
from .count_matrix import count_matrix_coo2_mult

__author__ = 'noe, marscher'


def _split_sequences_singletraj(dtraj, nstates, lag):
    """ splits the discrete trajectory into conditional sequences by starting state

    Parameters
    ----------
    dtraj : int-iterable
        discrete trajectory
    nstates : int
        total number of discrete states
    lag : int
        lag time

    """
    sall = [[] for _ in range(nstates)]
    res_states = []
    res_seqs = []

    for t in range(len(dtraj)-lag):
        sall[dtraj[t]].append(dtraj[t+lag])

    for i in range(nstates):
        if len(sall[i]) > 0:
            res_states.append(i)
            res_seqs.append(np.array(sall[i]))
    return res_states, res_seqs


def _split_sequences_multitraj(dtrajs, lag):
    """ splits the discrete trajectories into conditional sequences by starting state

    Parameters
    ----------
    dtrajs : list of int-iterables
        discrete trajectories
    nstates : int
        total number of discrete states
    lag : int
        lag time

    """
    from deeptime.markov import number_of_states
    n = number_of_states(dtrajs)
    res = []
    for i in range(n):
        res.append([])
    for dtraj in dtrajs:
        states, seqs = _split_sequences_singletraj(dtraj, n, lag)
        for i in range(len(states)):
            res[states[i]].append(seqs[i])
    return res


def _indicator_multitraj(ss, i, j):
    """ Returns conditional sequence for transition i -> j given all conditional sequences """
    iseqs = ss[i]
    res = []
    for iseq in iseqs:
        x = np.zeros(len(iseq))
        I = np.where(iseq == j)
        x[I] = 1.0
        res.append(x)
    return res


def _wrapper(*args):
    with threadpool_limits(limits=1, user_api='blas'):
        # writes results of statistical_inefficiency to destination memmap (array_fn)
        seqs, truncate_acf, mact, I, J, array_fn, start, stop = args[0]
        array = np.memmap(array_fn, mode='r+', dtype=np.float64)
        partial = np.empty(len(I))
        for n, (i, j) in enumerate(zip(I, J)):
             s = _indicator_multitraj(seqs, i, j)
             partial[n] = statistical_inefficiency(s, truncate_acf=truncate_acf, mact=mact)
        array[start:stop] = partial


class _arguments_generator:
    # splits I and J in blocks to serve n_jobs processes.
    def __init__(self, I, J, splitted_seqs, truncate_acf, mact, array, njobs):
        self.I = I
        self.J = J
        self.splitted_seqs = splitted_seqs
        self.truncate_acf = truncate_acf
        self.mact = mact
        self.array = array
        self.njobs = njobs

    def __len__(self):
        return len(self.I)

    def n_blocks(self):
        k = 4
        n = max(len(self.I) // (k*self.njobs - 1), 1)
        return n

    def __iter__(self):
        def chunks(n):
            for i in range(0, len(self.I), n):
                yield self.I[i:i + n], self.J[i:i+n]

        start = 0
        for n, (I, J) in enumerate(chunks(self.n_blocks())):
            stop = start + len(I)
            yield (self.splitted_seqs, self.truncate_acf, self.mact, I, J, self.array, start, stop)
            start = stop


def statistical_inefficiencies(dtrajs, lag, C=None, truncate_acf=True, mact=2.0, n_jobs=1, callback=None):
    r""" Computes statistical inefficiencies of sliding-window transition counts at given lag

    Consider a discrete trajectory :math`{ x_t }` with :math:`x_t \in {1, ..., n}`. For each starting state :math:`i`,
    we collect the target sequence

    .. mathh:
        Y^(i) = {x_{t+\tau} | x_{t}=i}

    which contains the time-ordered target states at times :math:`t+\tau` whenever we started in state :math:`i`
    at time :math:`t`. Then we define the indicator sequence:

    .. math:
        a^{(i,j)}_t (\tau) = 1(Y^(i)_t = j)

    The statistical inefficiency for transition counts :math:`c_{ij}(tau)` is computed as the statistical inefficiency
    of the sequence :math:`a^{(i,j)}_t (\tau)`.

    Parameters
    ----------
    dtrajs : list of int-iterables
        discrete trajectories
    lag : int
        lag time
    C : scipy sparse matrix (n, n) or None
        sliding window count matrix, if already available
    truncate_acf : bool, optional, default=True
        When the normalized autocorrelation function passes through 0, it is truncated in order to avoid integrating
        random noise
    n_jobs: int, default=1
        If greater one, the function will be evaluated with multiple processes.
    callback: callable, default=None
        will be called for every statistical inefficiency computed (number of nonzero elements in count matrix).
        If n_jobs is greater one, the callback will be invoked per finished batch.

    Returns
    -------
    I : scipy sparse matrix (n, n)
        Statistical inefficiency matrix with a sparsity pattern identical to the sliding-window count matrix at the
        same lag time. Will contain a statistical inefficiency :math:`I_{ij} \in (0,1]` whenever there is a count
        :math:`c_{ij} > 0`. When there is no transition count (:math:`c_{ij} = 0`), the statistical inefficiency is 0.

    See also
    --------
    deeptime.markov.tools.util.statistics.statistical_inefficiency
        used to compute the statistical inefficiency for conditional trajectories

    """
    import os
    # enforce one job on windows because of memmap file (not supported from multiple jobs)
    if os.name == 'nt':
        n_jobs = 1

    # count matrix
    if C is None:
        C = count_matrix_coo2_mult(dtrajs, lag, sliding=True, sparse=True)
    if callback is not None:
        if not callable(callback):
            raise ValueError('Provided callback is not callable')
    # split sequences
    splitseq = _split_sequences_multitraj(dtrajs, lag)
    # compute inefficiencies
    I, J = C.nonzero()
    if n_jobs > 1:
        from multiprocessing.pool import Pool
        import tempfile

        # to avoid pickling partial results, we store these in a numpy.memmap
        ntf = tempfile.NamedTemporaryFile(delete=False)
        arr = np.memmap(ntf.name, dtype=np.float64, mode='w+', shape=C.nnz)
        #arr[:] = np.nan
        gen = _arguments_generator(I, J, splitseq, truncate_acf=truncate_acf, mact=truncate_acf,
                                   array=ntf.name, njobs=n_jobs)
        if callback:
            x = gen.n_blocks()
            _callback = lambda _: callback(x)
        else:
            _callback = callback

        from deeptime.util.parallel import joining
        with joining(Pool(n_jobs)) as pool:
            result_async = [pool.apply_async(_wrapper, (args,), callback=_callback)
                            for args in gen]

            [t.get() for t in result_async]
            data = np.array(arr[:])
        import os
        os.unlink(ntf.name)
    else:
        data = np.empty(C.nnz)
        for index, (i, j) in enumerate(zip(I, J)):
            data[index] = statistical_inefficiency(_indicator_multitraj(splitseq, i, j),
                                                   truncate_acf=truncate_acf, mact=mact)
            if callback is not None:
                callback(1)
    res = csr_matrix((data, (I, J)), shape=C.shape)
    return res


def effective_count_matrix(dtrajs, lag, average='row', truncate_acf=True, mact=1.0, n_jobs=1, callback=None):
    r""" Computes the statistically effective transition count matrix

    Given a list of discrete trajectories, compute the effective number of statistically uncorrelated transition
    counts at the given lag time. First computes the full sliding-window counts :math:`c_{ij}(tau)`. Then uses
    :func:`statistical_inefficiencies` to compute statistical inefficiencies :math:`I_{ij}(tau)`. The number of
    effective counts in a row is then computed as

    .. math:
        c_i^{\mathrm{eff}}(tau) = \sum_j I_{ij}(tau) c_{ij}(tau)

    and the effective transition counts are obtained by scaling the rows accordingly:

    .. math:
        c_{ij}^{\mathrm{eff}}(tau) = \frac{c_i^{\mathrm{eff}}(tau)}{c_i(tau)} c_{ij}(tau)

    This procedure is not yet published, but a manuscript is in preparation [1]_.

    Parameters
    ----------
    dtrajs : list of int-ndarrays
        discrete trajectories
    lag : int
        lag time
    average : str, default='row'
        Use either of 'row', 'all', 'none', with the following consequences:
        'none': the statistical inefficiency is applied separately to each
            transition count (not recommended)
        'row': the statistical inefficiency is averaged (weighted) by row
            (recommended).
        'all': the statistical inefficiency is averaged (weighted) over all
            transition counts (not recommended).
    truncate_acf : bool, optional, default=True
        Mode of estimating the autocorrelation time of transition counts.
        True: When the normalized autocorrelation function passes through 0,
        it is truncated in order to avoid integrating random noise. This
        tends to lead to a slight underestimate of the autocorrelation time
    mact : float, default=2.0
        multiplier for the autocorrelation time. We tend to underestimate the
        autocorrelation time (and thus overestimate effective counts)
        because the autocorrelation function is truncated when it passes
        through 0 in order to avoid numerical instabilities.
        This is a purely heuristic factor trying to compensate this effect.
        This parameter might be removed in the future when a more robust
        estimation method of the autocorrelation time is used.

    n_jobs: int, default=1
        If greater one, the function will be evaluated with multiple processes.

    callback: callable, default=None
        will be called for every statistical inefficiency computed (number of nonzero elements in count matrix).
        If n_jobs is greater one, the callback will be invoked per finished batch.

    See also
    --------
    statistical_inefficiencies
        is used for computing the statistical inefficiences of sliding window
        transition counts

    References
    ----------
    .. [1] Noe, F. and H. Wu: in preparation (2015)

    """
    # observed C
    C = count_matrix_coo2_mult(dtrajs, lag, sliding=True, sparse=True)
    # statistical inefficiencies
    si = statistical_inefficiencies(dtrajs, lag, C=C, truncate_acf=truncate_acf, mact=mact,
                                    n_jobs=n_jobs, callback=callback)
    # effective element-wise counts
    Ceff = C.multiply(si)
    # averaging
    if average.lower() == 'row':
        # reduction factor by row
        factor = np.array(Ceff.sum(axis=1) / np.maximum(1.0, C.sum(axis=1)))
        # row-based effective counts
        Ceff = scipy.sparse.csr_matrix(C.multiply(factor))
    elif average.lower() == 'all':
        # reduction factor by all
        factor = Ceff.sum() / C.sum()
        Ceff = scipy.sparse.csr_matrix(C.multiply(factor))
    # else: by element, we're done.

    return Ceff



r"""This module implements the countmatrix estimation functionality

.. moduleauthor:: B.Trendelkamp-Schroer <benjamin DOT trendelkamp-schroer AT fu-berlin DOT de>
"""

import numpy as np
import scipy.sparse


################################################################################
# count_matrix
################################################################################

def count_matrix_coo2_mult(dtrajs, lag, sliding=True, sparse=True, nstates=None):
    r"""Generate a count matrix from a given list discrete trajectories.

    The generated count matrix is a sparse matrix in compressed
    sparse row (CSR) or numpy ndarray format.

    Parameters
    ----------
    dtraj : list of ndarrays
        discrete trajectories
    lag : int
        Lagtime in trajectory steps
    sliding : bool, optional
        If true the sliding window approach
        is used for transition counting
    sparse : bool (optional)
        Whether to return a dense or a sparse matrix
    nstates : int, optional
        Enforce a count-matrix with shape=(nstates, nstates). If there are
        more states in the data, this will lead to an exception.

    Returns
    -------
    C : scipy.sparse.csr_matrix or numpy.ndarray
        The countmatrix at given lag in scipy compressed sparse row
        or numpy ndarray format.

    """
    # Determine number of states
    if nstates is None:
        from deeptime.markov import number_of_states
        nstates = number_of_states(dtrajs)
    rows = []
    cols = []
    # collect transition index pairs
    for dtraj in dtrajs:
        if dtraj.size > lag:
            if sliding:
                rows.append(dtraj[0:-lag])
                cols.append(dtraj[lag:])
            else:
                rows.append(dtraj[0:-lag:lag])
                cols.append(dtraj[lag::lag])
    # is there anything?
    if len(rows) == 0:
        raise ValueError('No counts found - lag ' + str(lag) + ' may exceed all trajectory lengths.')
    # feed into one COO matrix
    row = np.concatenate(rows)
    col = np.concatenate(cols)
    data = np.ones(row.size)
    C = scipy.sparse.coo_matrix((data, (row, col)), shape=(nstates, nstates))
    # export to output format
    if sparse:
        return C.tocsr()
    else:
        return C.toarray()



r"""Dense implementation of hitting probabilities

.. moduleauthor:: F.Noe <frank DOT noe AT fu-berlin DOT de>

"""

import numpy as np


def hitting_probability(P, target):
    r"""
    Computes the hitting probabilities for all states to the target states.

    The hitting probability of state i to set A is defined as the minimal,
    non-negative solution of:

    .. math::
        h_i^A &= 1                    \:\:\:\:  i\in A \\
        h_i^A &= \sum_j p_{ij} h_i^A  \:\:\:\:  i \notin A

    Returns
    =======
    h : ndarray(n)
        a vector with hitting probabilities
    """
    if hasattr(target, "__len__"):
        target = np.array(target)
    else:
        target = np.array([target])
    # target size
    n = np.shape(P)[0]
    # nontarget
    nontarget = np.array(list(set(range(n)) - set(target)), dtype=int)
    # stable states
    stable = np.where(np.isclose(np.diag(P), 1) == True)[0]
    # everything else
    origin = np.array(list(set(nontarget) - set(stable)), dtype=int)
    # solve hitting probability problem (P-I)x = -b
    A = P[origin, :][:, origin] - np.eye((len(origin)))
    b = np.sum(-P[origin, :][:, target], axis=1)
    x = np.linalg.solve(A, b)
    # fill up full solution with 0's for stable states and 1's for target
    xfull = np.ones((n))
    xfull[origin] = x
    xfull[target] = 1
    xfull[stable] = 0

    return xfull



import itertools
from typing import Tuple, List, Optional

import numpy as np

from deeptime.util.decorators import plotting_function

Coordinate = Tuple[int, int]


class DrunkardsWalk:
    r""" This example dataset simulates the steps a drunkard living in a two-dimensional plane takes finding
    either the bar or the home as two absorbing states.

    The drunkard can take steps in a 3x3 stencil with uniform probability (as possible, in the corners the only
    possibilities are the ones that do not lead out of the grid). The transition matrix
    :math:`P\in\mathbb{R}^{nm\times nm}`  possesses one absorbing state for home and bar, respectively,
    and uniform two-dimensional jump probabilities in between. The grid is of size :math:`n\times m` and a point
    :math:`(i,j)` is identified with state :math:`i+nj` in the transition matrix.

    Parameters
    ----------
    grid_size : tuple
        The grid size, must be tuple of length two.
    bar_location : tuple
        The bar location, must be valid coordinate and tuple of length two.
    home_location : tuple
        The home location, must be valid coordinate and tuple of length two.
    barriers : List of tuple of two integers or None, default=None
        Initial barrier locations. Can also be added post-hoc by calling :meth:`add_barrier`.
    """

    def __init__(self, grid_size: Tuple[int, int], bar_location: List[Coordinate], home_location: List[Coordinate],
                 barriers=None):
        if barriers is None:
            barriers = []
        self.n_states = grid_size[0] * grid_size[1]
        self.grid_size = grid_size
        self.bar_location = bar_location if isinstance(bar_location, (tuple, list, np.ndarray)) else [bar_location]
        self.bar_location = np.atleast_2d(self.bar_location)
        self.bar_state = [self.coordinate_to_state(state) for state in self.bar_location]
        self.home_location = home_location if isinstance(home_location, (tuple, list, np.ndarray)) else [home_location]
        self.home_location = np.atleast_2d(self.home_location)
        self.home_state = [self.coordinate_to_state(state) for state in self.home_location]
        self.barriers = barriers
        self.barrier_states = [self.coordinate_to_state(barrier) for barrier in self.barriers]
        self.barrier_weights = [None]*len(barriers)

        from deeptime.markov.msm import MarkovStateModel
        self._msm = MarkovStateModel(transition_matrix=np.eye(self.n_states, dtype=np.float64))
        self._update_transition_matrix()

    def _is_home_or_bar_state(self, state):
        return state in self.home_state or state in self.bar_state

    def _is_home_or_bar_coord(self, coord):
        return self._is_home_or_bar_state(self.coordinate_to_state(coord))

    def _update_transition_matrix(self) -> None:
        r"""Updates the MSM so that the state of the simulator is reflected in the transition matrix.
        """
        transition_matrix = self.msm.transition_matrix.copy()
        for i in range(self.grid_size[0]):
            for j in range(self.grid_size[1]):
                coord = (i, j)
                state = self.coordinate_to_state(coord)  # row in the transition matrix
                next_steps = []
                for offset_i in [-1, 0, 1]:
                    for offset_j in [-1, 0, 1]:
                        if offset_i != 0 or offset_j != 0:
                            next_step = (coord[0] + offset_i, coord[1] + offset_j)
                            if self.is_valid_coordinate(next_step):
                                next_steps.append(next_step)
                #  uniform probability
                probabilities = []
                for next_step in next_steps:
                    next_state = self.coordinate_to_state(next_step)
                    if self.barriers is not None and next_step in self.barriers:
                        barrier_ix = self.barriers.index(next_step)
                        weight = self.barrier_weights[barrier_ix]
                        if weight is None:
                            probabilities.append(0.)
                        else:
                            probabilities.append(1./weight)
                    else:
                        if self._is_home_or_bar_state(state) and self._is_home_or_bar_state(next_state):
                            probabilities.append(100.)
                        else:
                            probabilities.append(1.)
                if self._is_home_or_bar_state(state) or \
                        (state in self.barrier_states and
                         self.barrier_weights[self.barrier_states.index(state)] is None):
                    # high probability to stay in home/bar
                    next_steps.append(coord)
                    probabilities.append(100.)
                else:
                    next_steps.append(coord)
                    probabilities.append(0.)
                probabilities = np.array(probabilities) / np.sum(probabilities)
                for p, step in zip(probabilities, next_steps):
                    transition_matrix[state, self.coordinate_to_state(step)] = p
        self.msm.update_transition_matrix(transition_matrix)

    def add_barrier(self, begin: Coordinate, end: Coordinate, weight: Optional[float] = None):
        r""" Adds a barrier to the grid by assigning probabilities

        .. math::
            P_{ij} = \mathbb{P}(X_{n+1} = j\in\mathrm{barriers} : X_n=i\text{ next to barrier}) = 0.

        The barrier is interpreted as a straight line between begin and end, discretized onto states using
        Bresenham's line algorithm :footcite:`bresenham1965algorithm`.

        Parameters
        ----------
        begin : tuple of two integers
            Begin coordinate of the barrier.
        end : tuple of two integers
            End coordinate of the barrier.
        weight : float or None, default=None
            If not None, the probability of jumping onto the barrier is set to :math:`1/w` before row-normalization.

        References
        ----------
        .. footbibliography::
        """
        barrier = []
        x0, y0 = begin
        x1, y1 = end

        dx = abs(x1 - x0)
        sx = 1 if x0 < x1 else -1
        dy = -abs(y1 - y0)
        sy = 1 if y0 < y1 else -1
        err = dx + dy

        while True:
            barrier.append((x0, y0))
            if x0 == x1 and y0 == y1:
                break
            e2 = 2*err
            if e2 >= dy:
                err += dy
                x0 += sx
            if e2 <= dx:
                err += dx
                y0 += sy
            if dy <= e2 <= dx:
                # introduced a 'diagonal' jump, append block so that diagonal is 'filled'
                barrier.append((x0, y0 - sy))

        self.barriers = list(itertools.chain(self.barriers, barrier))
        self.barrier_weights = list(itertools.chain(self.barrier_weights, [weight]*len(barrier)))
        self.barrier_states = [self.coordinate_to_state(barrier) for barrier in self.barriers]
        self._update_transition_matrix()

    def coordinate_to_state(self, coord: Coordinate) -> int:
        r""" Transforms a two-dimensional grid point (i, j) to a one-dimensional state.

        Parameters
        ----------
        coord : (i, j) tuple
            The grid point.

        Returns
        -------
        state : int
            The state corresponding to the grid point.
        """
        return coord[0] + self.grid_size[0] * coord[1]

    def state_to_coordinate(self, state: int) -> Coordinate:
        r""" Inverse operation to :meth:`coordinate_to_state`. Transforms state to corresponding coordinate (i,j).

        Parameters
        ----------
        state : int
            The state.

        Returns
        -------
        coordinate : (i, j) tuple
            The corresponding coordinate.
        """
        coord_j = np.floor(state / (self.grid_size[0]))
        coord_i = state - coord_j * self.grid_size[0]
        return coord_i, coord_j

    def is_valid_coordinate(self, coord: Coordinate) -> bool:
        r""" Validates if a coordinate is within bounds.

        Parameters
        ----------
        coord : (i, j) tuple
            The coordinate.

        Returns
        -------
        is_valid : bool
            Whether the coordinate is within bounds.
        """
        return (0 <= coord[0] < self.grid_size[0]) and (0 <= coord[1] < self.grid_size[1])

    @property
    def msm(self):
        r""" Yields a :class:`MSM <deeptime.markov.msm.MarkovStateModel>` which is parameterized with a transition matrix
        corresponding to this setup.

        Returns
        -------
        msm : deeptime.markov.msm.MarkovStateModel
            The corresponding Markov state model.
        """
        return self._msm

    def walk(self, start: Coordinate, n_steps: int, stop: bool = True, return_states: bool = False, seed: int = -1):
        r""" Simulates a random walk on the grid.

        Parameters
        ----------
        start : (i, j) tuple
            Start coordinate on the grid.
        n_steps : int
            Maximum number of steps to simulate.
        stop : bool, default=False
            Whether to stop the simulation once home or bar have been reached
        return_states : bool, default=False
            Whether to return states instead of coordinates. Default is False.
        seed : int, default=-1
            Random seed.

        Returns
        -------
        random_walk : (n_steps, 2) ndarray
            A random walk in coordinate space.
        """
        assert self.is_valid_coordinate(start), "Start must be within bounds."
        if stop:
            stopping_states = np.concatenate([self.home_state, self.bar_state])
        else:
            stopping_states = None
        states = self.msm.simulate(n_steps, start=self.coordinate_to_state(start), stop=stopping_states, seed=seed)
        if not return_states:
            return np.array([self.state_to_coordinate(state) for state in states])
        else:
            return states

    @staticmethod
    @plotting_function
    def plot_path(ax, path, intermediates: bool = True, color_lerp: bool = True, **plot_kw):  # pragma: no cover
        r""" Plots a path onto a drunkard's walk map. The path is interpolated with splines and, if desired, has a
        linearly interpolated color along its path.

        Parameters
        ----------
        ax : matplotlib axis
            The axis to plot onto.
        path : (N, 2) ndarray
            The path in terms of 2-dimensional coordinates.
        intermediates : bool, default=True
            Whether to scatterplot the states that were visited in the path.
        color_lerp : bool, default=True
            Whether to perform a linear interpolation along a colormap while the path progresses through the map.
        plot_kw
            Additional args that go into either a `ax.plot` call (`lerp=False`) or into a `LineCollection(...)` call
            (`lerp=True`).
        """
        import scipy.interpolate as interp
        from matplotlib.collections import LineCollection

        path = np.asarray(path)

        x = np.r_[path[:, 0]]
        y = np.r_[path[:, 1]]
        f, u = interp.splprep([x, y], s=0, per=False)
        xint, yint = interp.splev(np.linspace(0, 1, 50000), f)
        if intermediates:
            ax.scatter(x, y, label='Visited intermediates')

        if color_lerp:
            points = np.stack([xint, yint]).T.reshape(-1, 1, 2)
            segments = np.concatenate([points[:-1], points[1:]], axis=1)
            coll = LineCollection(segments, cmap='cool', linestyle='dotted', **plot_kw)
            coll.set_array(np.linspace(0, 1, num=len(points), endpoint=True))
            coll.set_linewidth(2)
            ax.add_collection(coll)
        else:
            ax.plot(xint, yint, **plot_kw)

    @plotting_function
    def plot_network(self, ax, F, cmap=None, connection_threshold: float = 0.):  # pragma: no cover
        r""" Plots a flux network onto a matplotlib axis assuming that the states are ordered according to a iota
        range (i.e., {0, 1, ..., n_states-1}) on the grid.

        Parameters
        ----------
        ax : matplotlib axis
            The axis to plot onto.
        F : (n_states, n_states) ndarray
            The flux network, assigning to a state tuple `F[i,j]` a current.
        cmap : matplotlib colormap, default=None
            A colormap to use for the edges.
        connection_threshold : float, default=0
            A connection threshold in terms of minimal current to exceed to draw an edge.

        Returns
        -------
        edge_colors : list
            List of edge colors.
        """
        import networkx as nx
        import matplotlib.pyplot as plt

        if cmap is None:
            cmap = plt.cm.jet

        G = nx.DiGraph()
        for state in range(self.n_states):
            G.add_node(state)

        edge_colors = []
        for state in range(self.n_states):
            for other_state in range(self.n_states):
                w = F[state, other_state]
                if w > connection_threshold:
                    G.add_edge(state, other_state, weight=w)
                    edge_colors.append(F[state, other_state])

        positions = np.zeros((self.n_states, 2))
        for i in range(self.n_states):
            positions[i] = self.state_to_coordinate(i)
        layout = {k: positions[k] for k in G.nodes.keys()}

        nx.draw_networkx_nodes(G, layout, node_size=30, ax=ax)
        nx.draw_networkx_edges(G, layout, ax=ax, arrowstyle='->', edge_cmap=cmap, edge_color=edge_colors,
                               connectionstyle='arc3, rad=0.1')
        return edge_colors

    @plotting_function
    def plot_2d_map(self, ax, barriers: bool = True, barrier_mode: str = 'filled'):  # pragma: no cover
        r""" Plots the drunkard's map onto an already existing matplotlib axis. One can select whether to draw barriers
        and if barriers should be drawn, whether they have a 'filled' face or be 'hollow' with just the border
        being colored.

        Parameters
        ----------
        ax : matplotlib axis
            The axis.
        barriers : bool, default=True
            Whether to draw the barriers.
        barrier_mode : str, default='filled'
            How to draw barriers - can be one of 'filled' and 'hollow'.

        Returns
        -------
        handles, labels
            Handles and labels for a matplotlib legend.
        """
        import numpy as np
        from matplotlib.patches import Rectangle

        valid_barrier_modes = ('filled', 'hollow')
        if barrier_mode not in valid_barrier_modes:
            raise ValueError(f"Barrier mode '{barrier_mode}' not supported, valid modes are {valid_barrier_modes}.")

        ax.scatter(*self.home_location.T, marker='*', label='Home', c='red', s=150, zorder=5)
        ax.scatter(*self.bar_location.T, marker='*', label='Bar', c='orange', s=150, zorder=5)

        ax.set_xticks(np.arange(10))
        ax.set_yticks(np.arange(10))
        ax.set_xlabel('coordinate x')
        ax.set_ylabel('coordinate y')

        rect = None
        for state in range(self.n_states):
            coord = self.state_to_coordinate(state)
            if self._is_home_or_bar_coord(coord):
                ax.add_patch(Rectangle((coord[0] - .5, coord[1] - .5), 1., 1., alpha=.3, color='green'))
            elif barriers and coord in self.barriers:
                barrier_ix = self.barriers.index(coord)
                weight = self.barrier_weights[barrier_ix]
                if weight is None:
                    rect = Rectangle((coord[0] - .5, coord[1] - .5), 1., 1., alpha=.5, color='red', lw=3.)
                    if barrier_mode == 'hollow':
                        rect.set_facecolor('none')
                    ax.add_patch(rect)
                else:
                    rect = Rectangle((coord[0] - .5, coord[1] - .5), 1., 1., alpha=.2, color='red', lw=3.)
                    if barrier_mode == 'hollow':
                        rect.set_facecolor('none')
                    ax.add_patch(rect)

        for grid_point in np.arange(-.5, self.grid_size[0] + .5, 1):
            ax.axhline(grid_point, linestyle='-', color='grey', lw=.5)

        for grid_point in np.arange(-.5, self.grid_size[1] + .5, 1):
            ax.axvline(grid_point, linestyle='-', color='grey', lw=.5)

        handles, labels = ax.get_legend_handles_labels()
        if rect is not None:
            handles.append(rect)
            labels.append("Barrier")

        ax.set_xlim([-.5, self.grid_size[0] - .5])
        ax.set_ylim([-.5, self.grid_size[1] - .5])

        return handles, labels



from typing import Tuple, Any, Dict

from . import _clustering_bindings as _bd


class MetricRegistry:
    r""" Registry of available metrics. Per default this contains only the Euclidean metric.
    If a custom metric is implemented, it can be registered through a call to
    :meth:`register <deeptime.clustering.MetricRegistry.register>`.

    Note that the registry should not be instantiated directly but rather be accessed
    through :data:`metrics <deeptime.clustering.metrics>`.
    """

    def __init__(self):
        self._registered = None
        self.register("euclidean", _bd.EuclideanMetric)

    def register(self, name: str, clazz):
        r""" Adds a new metric to the registry.

        Parameters
        ----------
        name : str
            The name of the metric.
        clazz : class
            Reference to the class of the metric.
        """
        self._mapping[name] = clazz

    @property
    def available(self) -> Tuple[str]:
        r"""List of registered metrics."""
        return tuple(self._mapping.keys())

    @property
    def _mapping(self) -> Dict[str, Any]:
        if self._registered is None:
            self._registered = dict()
        return self._registered

    def __getitem__(self, item: str):
        return self._mapping[item]


r""" Singleton instance of metrics registry. """
metrics = MetricRegistry()
